import warnings


class URScript(object):
    """
    Object for generating URScript. Adapted from python-urx.
    """

    def __init__(self, **kwargs):
        self.header = None
        self.program = None

        self.controller_ports = kwargs.get("controller_ports", [0, 1])
        self.controller_voltage = kwargs.get("controller_voltage", [0, 2])
        self.tool_ports = kwargs.get("tool_ports", [2, 3])
        self.tool_voltage = kwargs.get("tool_voltage", [0, 1, 2])
        self.output_domain_voltage = kwargs.get("output_domain_voltage", [0, 1])

    def __call__(self):
        if self.program is None:
            warnings.warn("URScript program is empty")
            return None

        # Construct the program
        myprog = """def myProg():{}\nend""".format(self.program)

        # Construct the full script
        script = ""
        if self.header:
            script = "{}\n\n".format(self.header)
        script = "{}{}".format(script, myprog)
        return script

    def reset(self):
        self.__init__()

    def add_header_to_program(self, header_line):
        self.header = "{}\n{}".format(self.header, header_line)

    def add_line_to_program(self, new_line):
        self.program = "{}\n\t{}".format(self.program, new_line)

    def _constrain_unsigned_char(self, value):
        """
        Ensure that unsigned char values are constrained to between 0 and 255.
        """
        assert(isinstance(value, int))
        if value < 0:
            value = 0
        elif value > 255:
            value = 255
        return value

    def set_analog_input_range(self, port, vrange):
        if port in self.controller_ports:
            assert(vrange in self.controller_voltage)
        elif port in self.tool_ports:
            assert(vrange in self.tool_voltage)
        msg = "set_analog_inputrange({},{})".format(port, vrange)
        self.add_line_to_program(msg)

    def set_analog_output(self, input_id, signal_level):
        assert(input_id in [0, 1])
        assert(signal_level in [0, 1])
        msg = "set_analog_output({}, {})".format(input_id, signal_level)
        self.add_line_to_program(msg)

    def set_analog_outputdomain(self, port, domain):
        assert(domain in self.output_domain_voltage)
        msg = "set_analog_outputdomain({},{})".format(port, domain)
        self.add_line_to_program(msg)

    def set_payload(self, mass, cog=None):
        msg = "set_payload({}".format(mass)
        if cog:
            assert(len(cog) == 3)
            msg = "{},{}".format(msg, cog)
        msg = "{})".format(msg)
        self.add_line_to_program(msg)

    def set_runstate_outputs(self, outputs=None):
        if not outputs:
            outputs = []
        msg = "set_runstate_outputs({})".format(outputs)
        self.add_line_to_program(msg)

    def set_tool_voltage(self, voltage):
        assert(voltage in [0, 12, 24])
        msg = "set_tool_voltage({})".format(voltage)
        self.add_line_to_program(msg)

    def sleep(self, value):
        msg = "sleep({})".format(value)
        self.add_line_to_program(msg)

    def socket_close(self, socket_name):
        msg = "socket_close(\"{}\")".format(socket_name)
        self.add_line_to_program(msg)

    def socket_get_var(self, var, socket_name):
        msg = "socket_get_var(\"{}\",\"{}\")".format(var, socket_name)
        self.add_line_to_program(msg)
        self.sync()

    def socket_open(self, socket_host, socket_port, socket_name):
        msg = "socket_open(\"{}\",{},\"{}\")".format(socket_host,
                                                     socket_port,
                                                     socket_name)
        self.add_line_to_program(msg)

    def socket_read_byte_list(self, nbytes, socket_name):
        msg = "global var_value = socket_read_byte_list({},\"{}\")".format(nbytes, socket_name)  # noqa
        self.add_line_to_program(msg)
        self.sync()

    def socket_send_string(self, message, socket_name):
        msg = "socket_send_string(\"{}\",\"{}\")".format(message, socket_name)  # noqa
        self.add_line_to_program(msg)
        self.sync()

    def socket_set_var(self, var, value, socket_name):
        msg = "socket_set_var(\"{}\",{},\"{}\")".format(var, value, socket_name)  # noqa
        self.add_line_to_program(msg)
        self.sync()

    def sync(self):
        msg = "sync()"
        self.add_line_to_program(msg)
